/* Check well-defined and L-attributed
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                               Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/
/*
  A grammer called ECLR attributed grammer if the grammer satisfy
  these three conditions.

        (1) L-attributed grammer
	(2) Well Defined
	(3) expression uniqueness for each EC and for each state

  This file 'check.c' includes functions to check (1) and (2).
  These functions are called by read_semantic_rules() in readsem.c.
*/

#include <stdio.h>
#include "attribute.h"
#include "error.h"
#include "pure.h"
#include "symtab.h"

extern bucket *startval;			/* reader.c  */
extern production_rule *current_prule;		/* reader.c  */
extern semantic_rule **current_semantic_rules_p;/* reader.c  */
extern attribute_list *appeared_attributes;	/* readsem.c  */
extern char tmp_buffer[];			/* readsem.c  */

#if (defined(__STDC__) || defined(__GNUC__))
extern semantic_rule *search_rule_in_first_semantic_rule(attribute *);	/* readsem.c  */
extern void remove_rule_from_first_semantic_rule(semantic_rule *);/*readsem.c*/

#else	/* not ANSI C nor GNU C  */
extern semantic_rule *search_rule_in_first_semantic_rule(); /* readsem.c  */
extern void remove_rule_from_first_semantic_rule();	    /* readsem.c  */

#endif	/* not ANSI C nor GNU C  */



#if (defined(__STDC__) || defined(__GNUC__))
void check_start_symbol(void);
void check_well_defined(void);
void check_L_attributed(void);
void check_unused_local(void);
static void check_assigned(attribute_list *, int);
static void cannot_refer(int, attribute *);
static void preceding_reference(int, attribute *);
static attribute_list *is_lhs_attribute(attribute *);

#else	/* not ANSI C nor GNU C  */
void check_start_symbol();
void check_well_defined();
void check_L_attributed();
void check_unused_local();
static void check_assigned();
static void cannot_refer();
static void preceding_reference();
static attribute_list *is_lhs_attribute();

#endif	/* not ANSI C nor GNU C  */



/* check if start symbol has only valid attributes.  */

void
check_start_symbol()
{
  register pure_list *plist;
  
  if (startval == NULL)
    internal("start symbol is not desided yet. at check_start_symbol()");
  
  plist = startval->pure_list;
  while (plist)
    {
      if (plist->pure->class == INHERIT)
	recovery_noline("start symbol \"%s\" mustn't have inherited attribute \"%s\".", startval->tag, plist->pure->name);
      plist = plist->link;
    }
}



/* check well defined, which is one of the condition
   which ECLR attributed grammer must satisfy.  */

void
check_well_defined()
{
  register production_rule *prule;

  /* check if synthesized attribute of lhs symbol is assigned.  */
  check_assigned(current_prule->attribute_list, SYNTHESIZE);

  /* check if inherited attribute of each rhs symbol is assigned.  */
  prule = current_prule->next;
  while (prule)
    {
      check_assigned(prule->attribute_list, INHERIT);
      prule = prule->next;
    }
}



/* check assigned.  */

static
void
check_assigned(alist, class)
     attribute_list *alist;
     short class;
{
  register attribute *attrp;
  register semantic_rule *rule;

  while (alist)
    {
      attrp = alist->attribute;
      rule = search_rule_in_first_semantic_rule(attrp);

      if (attrp->pure->class == class)
	{
	  if (rule == NULL)
	    {
	      if (attrp->index == 1)
		tmp_buffer[0] = '\0';
	      else
		sprintf(tmp_buffer, "[%d]", attrp->index);
	      
	      recovery("attribute \"%s%s.%s\" should be assigned."
			  , attrp->tag, tmp_buffer, attrp->pure->name);
	    }
	}
      else if (rule)
	{
	  if (attrp->index == 1)
	    tmp_buffer[0] = '\0';
	  else
	    sprintf(tmp_buffer, "[%d]", attrp->index);

	  recovery_with_line(rule->lineno,"\"%s%s.%s\" is only for reference."
			     , attrp->tag, tmp_buffer, attrp->pure->name);
	  remove_rule_from_first_semantic_rule(rule);
	}

      alist = alist->link;
    }
}



/* check L-attributed, which is one of the condition
   which ECLR attributed grammer must satisfy.  */

void
check_L_attributed()
{
  register semantic_expression *exp;
  register semantic_rule *rule;
  register attribute *attrp;
  register short loffset;

  rule = *current_semantic_rules_p;

  while (rule)
    {
      loffset = rule->attribute->offset;
      exp = rule->expression;
      
      switch (rule->attribute->pure->class) 
	{
	case INHERIT:
	  while (exp)
	    {
	      attrp = exp->pointer.attribute;

	      switch (exp->kind)
		{
		case KINHERIT:
		  if (is_lhs_attribute(attrp) == NULL)
		    cannot_refer(rule->lineno, attrp);
		  break;

		case KSYNTHESIZE:
		  if (is_lhs_attribute(attrp))
		    cannot_refer(rule->lineno, attrp);
		  else if (loffset < attrp->offset)
		    preceding_reference(rule->lineno, attrp);
		  break;

		case KLOCAL:
		  if (loffset < attrp->offset)
		    preceding_reference(rule->lineno, attrp);
		  break;
		}

	      exp = exp->next;
	    }
	  break;

	case SYNTHESIZE:
	  while (exp)
	    {
	      attrp = exp->pointer.attribute;

	      switch (exp->kind)
		{
		case KINHERIT:
		  if (is_lhs_attribute(attrp) == NULL)
		    cannot_refer(rule->lineno, attrp);
		  break;

		case KSYNTHESIZE:
		  if (is_lhs_attribute(attrp))
		    cannot_refer(rule->lineno, attrp);
		  break;
		}
	      exp = exp->next;
	    }
	  break;

	case LOCAL:
	  while (exp)
	    {
	      attrp = exp->pointer.attribute;

	      switch (exp->kind)
		{
		case KINHERIT:
		  if (is_lhs_attribute(attrp) == NULL)
		    cannot_refer(rule->lineno, attrp);
		  break;

		case KSYNTHESIZE:
		  if (is_lhs_attribute(attrp))
		    cannot_refer(rule->lineno, attrp);
		  break;
		}
	      exp = exp->next;
	    }
	  break;
	}			/* switch  */

      rule = rule->link;
    }				/* while  */
}



/* print an error message about preceding reference.  */

static
void
preceding_reference(lno, attrp)
     int lno;
     attribute *attrp;
{
  register pure *pp;

  pp = attrp->pure;
  switch (pp->class)
    {
    case SYNTHESIZE:
      if (attrp->index == 1)
	tmp_buffer[0] = '\0';
      else
	sprintf(tmp_buffer, "[%d]", attrp->index);

      recovery_with_line(lno, "preceding reference \"%s%s.%s\"."
			    , attrp->tag, tmp_buffer, pp->name);
      break;

    case LOCAL:
      recovery_with_line(lno, "preceding reference \"%s\".", pp->name);
      break;
    }
}



/* print an error message that the attribute cannot refer.  */

static
void
cannot_refer(lno, attrp)
     int lno;
     attribute *attrp;
{
  register pure *pp;

  pp = attrp->pure;
  if (pp->class == INHERIT || pp->class == SYNTHESIZE)
    {
      if (attrp->index == 1)
	tmp_buffer[0] = '\0';
      else
	sprintf(tmp_buffer, "[%d]", attrp->index);
      recovery_with_line(lno, "cannot refer \"%s%s.%s\"."
			 , attrp->tag, tmp_buffer, pp->name);
    }
}



/* check if attribute is left hand side of production rule or not.  */

static
attribute_list *
is_lhs_attribute(attrp)
     attribute *attrp;
{
  register attribute_list *alist;

  alist = current_prule->attribute_list;

  while (alist)
    {
      if (alist->attribute == attrp)
	break;

      alist = alist->link;
    }

  return (alist);
}



/* check unused local attribute.  */

void
check_unused_local()
{
  register attribute_list *alist;
  register attribute *attrp;

  alist = appeared_attributes;

  while (alist)
    {
      attrp = alist->attribute;

      if (attrp->pure->class == LOCAL
	  &&
	  search_rule_in_first_semantic_rule(attrp) == NULL)
	warning("local attribute \"%s\" is unused.", attrp->pure->name);

      alist = alist->link;
    }
}
