/* Check equivalence classes
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

/*
  Check ECLR-ness for each state.

  A grammer called ECLR attributed grammer iif the grammer satisfy
  these three conditions.

  	(1) L-attributed grammer
	(2) Well Defined
	(3) expression uniqueness for each EC and for each state
	
  eclr() checks only (3).  (1) and (2) are already checked at
  readgram() in reader.c while reading grammers(production rules and 
  semantic rules).

  entry point is eclr().

  description of variables :

  'closed'	a bit set of nonterminals.  each bit indicates that
  		these has already closed by the nonterminal (bit is
		set) or not (bit is reset).
		check_eclr() only uses this.

  .....
*/

#include <stdio.h>
#include "attribute.h"
#include "error.h"
#include "files.h"
#include "gram.h"
#include "machine.h"
#include "new.h"
#include "pure.h"
#include "state.h"
#include "system.h"

#if (defined(__STDC__) || defined(__GNUC__))
extern void output_inherit_header(void);	/* outputsem.c  */
extern void output_inherit_trailer(void);	/* outputsem.c  */
extern void output_inherit_pre(int);		/* outputsem.c  */
extern void output_inherit_post(void);		/* outputsem.c  */
extern void output_inherit(semantic_rule **);	/* outputsem.c  */
extern void output_local(semantic_rule_list *);	/* outputsem.c  */
extern void print_EC_output(semantic_rule_list **);	/* printsem.c  */
extern void print_local_output(semantic_rule_list *);	/* printsem.c  */
extern void print_cores(short *, short *);		/* printsem.c  */
extern void print_terminator(void);			/* printsem.c  */

#ifdef DEBUG
extern void print_expressions(semantic_expression *);	/* attribute.c  */
extern void print_semrule_d(semantic_rule *);		/* attribute.c  */
extern void print_semrules_d(semantic_rule *);		/* attribute.c  */
#endif

#else	/* not ANSI C nor GNU C  */
extern void output_inherit_header();	/* outputsem.c  */
extern void output_inherit_trailer();	/* outputsem.c  */
extern void output_inherit_pre();	/* outputsem.c  */
extern void output_inherit_post();	/* outputsem.c  */
extern void output_inherit();		/* outputsem.c  */
extern void output_local();		/* outputsem.c  */
extern void print_EC_output();		/* printsem.c  */
extern void print_local_output();	/* printsem.c  */
extern void print_cores();		/* printsem.c  */
extern void print_terminator();		/* printsem.c  */

#ifdef DEBUG
extern void print_expressions();	/* attribute.c  */
extern void print_semrule_d();		/* attribute.c  */
extern void print_semrules_d();		/* attribute.c  */
#endif

#endif	/* not ANSI C nor GNU C  */     

extern int failure;			/* main.c  */
extern char **tags;			/* reader.c  */
extern short **derives;			/* derives.c  */
extern core **state_table;		/* lalr.c  */
extern int verboseflag;			/* getargs.c  */
extern int simple_verboseflag;		/* getargs.c  */
extern semantic_rule **inh_table;	/* attribute.c  */
extern semantic_rule **synt_table;	/* attribute.c  */
extern semantic_rule **local_table;	/* attribute.c  */
extern int nlocal_classes;		/* attribute.c  */
extern int nEC;				/* attribute.c  */

int nviolation;		/* number of ECLR violation  */


#if (defined(__STDC__) || defined(__GNUC__))

void eclr(void);
static void initialize_eclr(void);
static void finalize_eclr(void);
static void check_eclr(short *, int);
static void check_uniqueness_kernel(semantic_rule *);
static void check_uniqueness(semantic_rule *);
static void align_offset(semantic_rule *);
static semantic_rule *expand_local(semantic_rule *);
static semantic_rule *expand_IN(semantic_rule *);
static semantic_rule *expand_IN_align_local(semantic_rule *);
static semantic_rule *compare_expressions(semantic_rule *, semantic_rule *);
static semantic_rule_list *add_to_EC_output(semantic_rule *);
static void eclr_violation(int);

static void clear_EC(void);
static void clear_local(void);
static void free_semexps_pure(semantic_expression *);
static void free_semexp_pure(semantic_expression *);
static void free_semrule_pure(semantic_rule *);
static void free_semrules_pure(semantic_rule *);
static semantic_rule *copy_semrules_pure(semantic_rule *);
static semantic_expression *copy_semexps_pure(semantic_expression *);

#ifdef DEBUG
static void print_item(int);
static void print_EC_expanded(void);
#endif

#else	/* not ANSI C nor GNU C  */

void eclr();
static void initialize_eclr();
static void finalize_eclr();
static void check_eclr();
static void check_uniqueness_kernel();
static void check_uniqueness();
static void align_offset();
static semantic_rule *expand_local();
static semantic_rule *expand_IN();
static semantic_rule *expand_IN_align_local();
static semantic_rule *compare_expressions();
static semantic_rule_list *add_to_EC_output();
static void eclr_violation();

static void clear_EC();
static void clear_local();
static void free_semexps_pure();
static void free_semexp_pure();
static void free_semrule_pure();
static void free_semrules_pure();
static semantic_rule *copy_semrules_pure();
static semantic_expression *copy_semexps_pure();

#ifdef DEBUG
static void print_item();
static void print_EC_expanded();
#endif

#endif	/* not ANSI C nor GNU C  */


static unsigned *closed;
static short *itemset;
static short *itemsetend;
static semantic_rule **EC_to_write;
static semantic_rule **EC_expanded;
static semantic_rule **EC_rule;
static semantic_rule *INlist;
static semantic_rule_list *local_to_write;
static semantic_rule_list *local_to_write_last;
static semantic_rule_list *local_output;
static semantic_rule_list *local_output_last;
static semantic_rule_list **EC_output;
static semantic_rule_list **EC_output_tail;

/* check expression-uniqueness for all states.  */

void
eclr()
{
  register core **state, **stateend;
  register semantic_rule **tp, **tpend;
  register int last_violation, violation_states;
  
  /* no ECLR violations at first  */
  nviolation = 0;

  initialize_eclr();

  violation_states = 0;
  
  tpend = EC_to_write + nEC;
  
  /* -3 means final state, next-to-final state and unknown state.  */
  state = state_table;
  stateend = state + nstates - 3;

  /* output header into table file.  */
  if (nEC > 0 || nlocal_classes > 0)
    output_inherit_header();

  /* for all states  */
  while (state < stateend)
    {
#ifdef DEBUG
      printf("\n\nCheck ECLR-ness for state %d\n", (*state)->number);
#endif
      if (nlocal_classes > 0)
	clear_local();
      clear_EC();
      free_semrules_pure(INlist);
      INlist = NULL;
      last_violation = nviolation;
      
      /* check each equivalence classes and take closures.  */
      check_eclr((*state)->items, (*state)->nitems);

      /* is there some ECLR violation in this state ?  */
      if (last_violation != nviolation)
	violation_states++;

      /* is there at least one inherited attribution rule ?  */
      tp = EC_to_write;
      while (tp < tpend)
	{
	  if (*tp)
	    break;
	  tp++;
	}

      if (verboseflag && !simple_verboseflag)
	{
	  print_cores(itemset, itemsetend);
	  print_terminator();
	}

      if (tp < tpend || local_to_write)
	{
	  /* there is at least one rule.  */
	  output_inherit_pre((*state)->number);
	  if (nlocal_classes > 0)
	    output_local(local_to_write);
	  output_inherit(EC_to_write);
	  output_inherit_post();

	  if (verboseflag && !simple_verboseflag)
	    {
	      print_local_output(local_output);
	      print_EC_output(EC_output);
	    }
	}

      if (verboseflag && !simple_verboseflag)
	print_terminator();
      
      state++;
    }

  /* output trailer into table file.  */
  if (nEC > 0 || nlocal_classes > 0)
    output_inherit_trailer();
  
  finalize_eclr();

  /* are there some ECLR violations ?  */
  if (nviolation > 0)
    eclr_violation(violation_states);
}
    

/* allocate and initialize variables.  */

static
void
initialize_eclr()
{
  INlist = NULL;
  closed = NEW2(WORDSIZE(nvars), unsigned);
  itemset = NEW2(nrules, short);
  EC_to_write = NEW2(nEC, semantic_rule *);
  EC_expanded = NEW2(nEC, semantic_rule *);
  EC_rule = NEW2(nEC, semantic_rule *);
  local_to_write = local_to_write_last = NULL;
  if (verboseflag && !simple_verboseflag)
    {
      EC_output = NEW2(nEC, semantic_rule_list *);
      EC_output_tail = NEW2(nEC, semantic_rule_list *);
      local_output = local_output_last = NULL;
    }
}


/* free variables.  */

static
void
finalize_eclr()
{
  free_semrules_pure(INlist);
  FREE(closed);
  FREE(itemset);
  clear_EC();
  FREE(EC_to_write);
  FREE(EC_expanded);
  FREE(EC_rule);
  if (nlocal_classes > 0)
    clear_local();
  if (verboseflag && !simple_verboseflag)
    {
      FREE(EC_output);
      FREE(EC_output_tail);
    }
}



/*
  check if the state satisfy expression-uniquness by each equivalence class.
  'n' items of array 'kernel[]' are the kernel of the state.
  we check the expression-uniquness with taking closure from kernel items.
*/ 

static
void
check_eclr(kernel, n)
     short *kernel;
     int n;
{
  register short *kernelend;
  register short *item, *ip, symbol, *ruleno;
  register int i;
  register semantic_rule *rule, *rp;
  register semantic_rule_list *lp;
  
  item = itemsetend = itemset;

  /* clear 'closed'  */
  for (i=0; i < WORDSIZE(nvars); i++)
    closed[i] = 0;
  
  if (n == 0)
    {
      /* initial state  */
      SETBIT(closed, start_symbol - ntokens);
      ruleno = derives[start_symbol];
      while (*ruleno > 0)
	*itemsetend++ = rrhs[*ruleno++];
    }
  else
    {
      /* copy kernel items into 'itemset[]'.  */
      kernelend = kernel + n;
      while (kernel < kernelend)
	{
#ifdef DEBUG
	  printf("kernel item...\n\t");
	  print_item(*kernel);
#endif
	  if (ritem[*kernel] > 0)
	    {
	      /* inherited attributes of this symbol belong to IN  */
	      rule = inh_table[*kernel];
	      while (rule)
		{
		  check_uniqueness_kernel(rule);
		  rule = rule->link;
		}

	      /* enlist rules of local attribute. append it to the last.  */
	      rule = local_table[*kernel];
	      while (rule)
		{
		  rp = copy_semrules_pure(rule);
		  align_offset(rp);
		  lp = NEW(semantic_rule_list);
		  lp->rule = rp;
		  lp->link = NULL;
		  if (local_to_write)
		    local_to_write_last->link = lp;
		  else	/* the first  */
		    local_to_write = lp;
		  local_to_write_last = lp;
		  
		  /* internal("allocation of local attribute is wrong. at check_eclr()"); */
		  if (verboseflag && !simple_verboseflag)
		    {
		      lp = NEW(semantic_rule_list);
		      lp->rule = rule;
		      lp->link = NULL;
		      if (local_output)
			local_output_last->link = lp;
		      else	/* the first  */
			local_output = lp;
		      local_output_last = lp;
		    }
		  rule = rule->link;
		}
	    }
	  
	  /* copy kernel items into 'itemset[]'  */
	  *itemsetend++ = *kernel++;
	}			/* while  */
  
      /* take a closure derived from kernel items  */
      item = itemset;
      ip = itemsetend;
      while (item < ip)
	{
	  symbol = ritem[*item];
	  if (ISVAR(symbol) && ! BITISSET(closed, symbol-ntokens))
	    {
#ifdef DEBUG
	      printf("taking closures derived by \"%s\"\n", tags[symbol]);
#endif
	      SETBIT(closed, symbol-ntokens);
	      ruleno = derives[symbol];

	      while (*ruleno > 0)
#ifdef DEBUG
		{
		  printf("add to the closure ");
		  print_item(rrhs[*ruleno]);
#endif
		  *itemsetend++ = rrhs[*ruleno++];
#ifdef DEBUG
		}
#endif
	    }
	  item++;
	}
    }	/* if..else  */

  
  while (item < itemsetend)
    {
      /* check all semantic rules, which deside inherited attribute of IN,
	 whether they satisfy expression uniqueness on their equivalence
	 classes.  */
      
      symbol = ritem[*item];
      if (symbol > 0)
	{
	  rule = inh_table[*item];
	  while (rule)
	    {
	      check_uniqueness(rule);
	      rule = rule->link;
	    }

	  rule = local_table[*item];
	  while (rule)
	    {
	      rp = copy_semrules_pure(rule);
	      expand_IN_align_local(rp);
	      lp = NEW(semantic_rule_list);
	      lp->rule = rp;
	      lp->link = NULL;
	      if (local_to_write)
		local_to_write_last->link = lp;
	      else	/* the first  */
		local_to_write = lp;
	      local_to_write_last = lp;

	      /* internal("allocation of local attribute is wrong. at check_eclr()"); */
	      if (verboseflag && !simple_verboseflag)
		{
		  lp = NEW(semantic_rule_list);
		  lp->rule = rule;
		  lp->link = NULL;
		  if (local_output)
		    local_output_last->link = lp;
		  else	/* the first  */
		    local_output = lp;
		  local_output_last = lp;
		}
	      rule = rule->link;
	    }
      
	  /* if the symbol followed by the dot is a nonterminal, take
	     a closure derived from the nonterminal iff it has not yet.  */
	  if (ISVAR(symbol) && ! BITISSET(closed, symbol-ntokens))
	    {
#ifdef DEBUG
	      printf("taking closures derived by \"%s\"\n", tags[symbol]);
#endif
	      SETBIT(closed, symbol-ntokens);
	      ruleno = derives[symbol];

	      while (*ruleno > 0)
#ifdef DEBUG
		{
		  printf("add to the closure ");
		  print_item(rrhs[*ruleno]);
#endif
		  *itemsetend++ = rrhs[*ruleno++];
#ifdef DEBUG
		}
	      printf("current items...\n");
	      ip = itemset;
	      while (ip < itemsetend)
		print_item(*ip++);
#endif
	    }
	}
      
      item++;
    }
#ifdef DEBUG
  printf("final expressions for each EC are...\n");
  print_EC_expanded();
  printf("Checking ECLR-ness finished.\n");
#endif
}


/* check uniqueness for the rule whose production rule is in kernel. */

static
void
check_uniqueness_kernel(rule)
     semantic_rule *rule;
{
  register int ec;
  register semantic_rule *rp, *rp2;
  register semantic_rule_list *olist;
  
#ifdef DEBUG
  printf("\nCheck uniqueness(kernel) : ");
  print_semrule_d(rule);
#endif

  /* check only rules for inherited attribute  */
  if (rule->attribute->pure->class != INHERIT)
    return;
    
  ec = rule->attribute->pure->EC;
  rp = copy_semrules_pure(rule);
  expand_local(rp);
  align_offset(rp);

  /* enlist into 'INlist'  */
  rp2 = copy_semrules_pure(rule);
  align_offset(rp2);
  rp2->link = INlist;
  INlist = rp2;
#ifdef DEBUG
  printf("add to INlist : ");
  print_semrule_d(rp2);
  printf("INlist are...\n");
  print_semrules_d(INlist);
#endif

  if (verboseflag && !simple_verboseflag)
    olist = add_to_EC_output(rule);

  /* check expression uniqueness on the equivalence class  */
  if (EC_expanded[ec] == NULL)
    /* the expression for the equivalence class 'ec' has not been
       described yet.  */
    {
#ifdef DEBUG
      printf("Add to EC_expanded[%d] <= ", ec);
      print_semrule_d(rp);
#endif
      EC_expanded[ec] = rp;
      EC_to_write[ec] = rp2;
      EC_rule[ec] = rule;
    }

  else	/* EC_expanded[ec] != NULL  */
    {
      if (compare_expressions(EC_expanded[ec], rp) == NULL)
	/* ECLR violation. the expression 'expanded' is not the same
	   expression already described on the equivalence class.  */
	{
#ifdef DEBUG
	  printf("## ECLR violation between belows ##\n");
	  printf("EC_expanded[%d] : ", ec);
	  print_semrule_d(EC_expanded[ec]);
	  printf("rule expanded   : ");
	  print_semrule_d(rp);
#endif	  
	  nviolation++;
	  if (verboseflag && !simple_verboseflag)
	    olist->flag = ECLR_VIOLATION;
	}
      free_semrules_pure(rp);
    }

#ifdef DEBUG
  print_EC_expanded();
  printf("Check uniqueness(kernel) : FINISH\n\n");
#endif
}



/* check the expression-uniqueness of each equivalence classes.
   the expression is given by 'rule'.  */

static
void
check_uniqueness(rule)
     semantic_rule *rule;
{
  register int ec;
  register semantic_rule *rp, *rp2;
  register semantic_rule_list *olist;
  
#ifdef DEBUG
  printf("\nCheck uniqueness : ");
  print_semrule_d(rule);
#endif

  ec = rule->attribute->pure->EC;
  rp = copy_semrules_pure(rule);
  expand_local(rp);
  expand_IN(rp);

  /* enlist into 'INlist'  */
  rp2 = copy_semrules_pure(rule);
  expand_IN_align_local(rp2);
  rp2->link = INlist;
  INlist = rp2;
#ifdef DEBUG
  printf("add to INlist : ");
  print_semrule_d(rp2);
  printf("INlist are...\n");
  print_semrules_d(INlist);
#endif

  if (verboseflag && !simple_verboseflag)
    olist = add_to_EC_output(rule);

  /* check expression uniqueness on the equivalence class  */
  if (EC_expanded[ec] == NULL)
    /* the expression for the equivalence class 'ec' has not been
       described yet.  */
    {
#ifdef DEBUG
      printf("Add to EC_expanded[%d] = ", ec);
      print_semrule_d(rp);
#endif

      EC_expanded[ec] = rp;
      EC_to_write[ec] = rp2;
      EC_rule[ec] = rule;
    }

  else	/* EC_expanded[ec] != NULL  */
    {
      if (compare_expressions(EC_expanded[ec], rp) == NULL)
	/* ECLR violation. the expression 'expanded' is not the same
	   expression already described on the equivalence class.  */
	{
#ifdef DEBUG
	  printf("## ECLR violation between belows ##\n");
	  printf("EC_expanded[%d] : ", ec);
	  print_semrule_d(EC_expanded[ec]);
	  printf("rule expanded   : ");
	  print_semrule_d(rp);
#endif	  
	  nviolation++;
	  if (verboseflag && !simple_verboseflag)
	    olist->flag = ECLR_VIOLATION;
	}
      free_semrules_pure(rp);
    }
#ifdef DEBUG
  printf("Check uniqueness : FINISH\n\n");
  print_EC_expanded();
#endif
}


/* align each offset of expressions.
   you shoud know that the values of each 'pointer.attribute->offset'
   member is *actually* changed.  but the offet of left-hand-side
   attribute is never changed.  */

static
void
align_offset(rule)
     semantic_rule *rule;
{
  register short base;
  register semantic_expression *ep;
  
  base = rule->attribute->offset;
  ep = rule->expression;

  while (ep)
    {
      switch (ep->kind)
	{
	case KINHERIT:
	case KSYNTHESIZE:
	case KLOCAL:
	  ep->pointer.attribute->offset -= base;
	  break;

	default:
	  /* do nothing */
	  break;
	}
      ep = ep->next;
    }
#ifdef DEBUG
  printf("align offset : ");
  print_semrule_d(rule);
#endif
}


 
/* all local attributes appeared in the expressions are expanded into
   the no-local-attribute-included expressions.
   the expression of given rule is *actually* changed, if some
   expansion has been done.						*/

static
semantic_rule *
expand_local(rule)
     semantic_rule *rule;
{
  register semantic_expression *previous, *ep, *ep2;
  register semantic_rule **lp, **lpend, *rp;
  register short *sp;
  register semantic_expression *ep3;

  previous = NULL;
  ep = rule->expression;

  while (ep)
    {
      if (ep->kind == KLOCAL)
	{
	  sp = &ritem[ep->pointer.attribute->offset];
	  while (*sp > 0)
	    sp++;
	  lp = &local_table[rrhs[-*sp]];
	  lpend = &local_table[ep->pointer.attribute->offset];

	  /* search this local attribute  */
	  while (lp <= lpend)
	    {
	      rp = *lp++;
	      while (rp)
		{
		  if (rp->attribute->pure == ep->pointer.attribute->pure)
		    /* find  */
		    break;
		  rp = rp->link;
		}
	      if (rp)
		break;
	    }

	  if (rp == NULL)     /* also 'lp > lpend'  */
	    internal("at expand_local()");

	  /* really expand the local attribute  */
	  ep2 = copy_semexps_pure(rp->expression);
	  if (previous)
	    previous->next = ep2;
	  else
	    rule->expression = ep2;

	  ep3 = ep2;

	  while(ep2->next)
	    ep2 = ep2->next;

	  ep2->next = ep->next;
	  FREE(ep);                /* FREE EXPS PURE */
	  ep = ep3;
	}
      else      /* ep->kind != KLOCAL  */
	{
	  previous = ep;
	  ep = ep->next;
	}

    }   /* while  */
#ifdef DEBUG
  printf("expand local : ");
  print_semrule_d(rule);
#endif
  return (rule);
}



/* Expand inherited attributes, which belong to IN, into whose expression.
   Therefore, all expressions became to be codes or local attributes or
   attributes of DEF.  */

static
semantic_rule *
expand_IN(rule)
     semantic_rule *rule;
{
    register semantic_expression *previous, *ep, *ep2;
    register short ec;
    
    previous = NULL;
    ep = rule->expression;

    while (ep)
      {
	if (ep->kind == KINHERIT && ep->pointer.attribute->offset >= 0)
	  {
	    /* expand it */
	    ec = ep->pointer.attribute->pure->EC;

	    if (EC_expanded[ec] == NULL)
	      internal("at expand_IN()");

#ifdef DEBUG
	    printf("expanding \"%s.%s[%d]{%d}\" => ",
		   ep->pointer.attribute->tag,
		   ep->pointer.attribute->pure->name,
		   ep->pointer.attribute->offset, ec);
	    print_expressions(EC_expanded[ec]->expression);
#endif
	    
	    ep2 = copy_semexps_pure(EC_expanded[ec]->expression);

	    if (previous)
	      previous->next = ep2;
	    else
	      rule->expression = ep2;
		
	    while (ep2->next)
	      ep2 = ep2->next;
	    ep2->next = ep->next;
	    free_semexp_pure(ep);
	    ep = ep2;
	  }	/* if  */

	previous = ep;
	ep = ep->next;
      }		/* while */
#ifdef DEBUG
    printf("expand IN : ");
    print_semrule_d(rule);
#endif
    return (rule);
}


static
semantic_rule *
expand_IN_align_local(rule)
     semantic_rule *rule;
{
    register semantic_expression *previous, *ep, *ep2;
    register semantic_rule *rp;
    register char *tag, *name;
    register attribute *ap;
    register int base;
    
    previous = NULL;
    base = rule->attribute->offset;
    ep = rule->expression;
    
    while (ep)
      {
	switch (ep->kind)
	  {
	  case KINHERIT:
	    if (ep->pointer.attribute->offset >= 0)
	      {
		/* expand it */
		ap = ep->pointer.attribute;
		tag = ap->tag;
		name = ap->pure->name;

		/* search the definition of attribute in 'INlist'  */
		rp = INlist;
		while (rp)
		  {
		    /* IF YOU DON'T USE THE STRING MANAGER.
		       USE 'strcmp()' INSTEAD.  */
		    if (rp->attribute->tag == tag
			&&
			rp->attribute->pure->name == name)
		      /* found  */
		      break;

		    rp = rp->link;
		  }		
	    
		if (rp == NULL)
		  internal("at expand_IN_align_local()");

#ifdef DEBUG
		printf("expanding \"%s.%s[%d]{%d}\" => ",
		       ap->tag, ap->pure->name, ap->offset, ap->pure->EC);
		print_expressions(rp->expression);
#endif
		ep2 = copy_semexps_pure(rp->expression);
		if (previous)
		  previous->next = ep2;
		else
		  rule->expression = ep2;
		
		while (ep2->next)
		  ep2 = ep2->next;
		ep2->next = ep->next;
		free_semexp_pure(ep);
		ep = ep2;
	      }	/* if  */
	    break;

	  case KLOCAL:
	    ep->pointer.attribute->offset -= base;
	    break;
	  }	/* switch  */

	previous = ep;
	ep = ep->next;
      }		/* while */
#ifdef DEBUG
    printf("expand IN align local: ");
    print_semrule_d(rule);
#endif
    return (rule);
}



/* Compare two expressions given by both rules. if both are same, return
   the expression (perhaps non-NULL).  If these are diffent, return NULL.
   But if you give both arguments NULL, return NULL.  */

static
semantic_rule *
compare_expressions(rule1, rule2)
     semantic_rule *rule1;
     semantic_rule *rule2;
{
  register semantic_expression *e1, *e2;
  register attribute *attr1, *attr2;
  
#ifdef DEBUG
  printf("compare expressions between...\n");
  print_semrule_d(rule1);
  print_semrule_d(rule2);
#endif
  
  e1 = rule1->expression;
  e2 = rule2->expression;
  
  while (e1)
    {
      if (e2 == NULL || e1->kind != e2->kind)
	return (NULL);
      
      switch (e1->kind) 
	{
	case KCODE:
	  /* IF YOU DON'T USE THE STRING MANAGER. USE 'strcmp()' INSTEAD.  */
	  if (e1->pointer.code != e2->pointer.code)
	    return (NULL);
	  break;

	case KINHERIT:
	  attr1 = e1->pointer.attribute;
	  attr2 = e2->pointer.attribute;
	  if (attr1->offset != attr2->offset
	      ||
	      attr1->pure->EC != attr2->pure->EC)
	    return (NULL);
	  break;

	case KSYNTHESIZE:
	  if (e1->pointer.attribute->offset != e2->pointer.attribute->offset)
	    return (NULL);
	  break;

	default:
	  internal("at compare_expressions()");
	  break;
	}	  
      
      e1 = e1->next;
      e2 = e2->next;
    }	/* while  */

  /* if there leave some expressions.  */
  if (e2)
    return (NULL);

#ifdef DEBUG
  printf("... both above are same\n");
#endif
  return (rule1);
}



static
semantic_rule_list *
add_to_EC_output(rule)
     semantic_rule *rule;
{
  register semantic_rule_list *listp;
  register short ec;

  ec = rule->attribute->pure->EC;
  
  listp = NEW(semantic_rule_list);
  listp->flag = ECLR_OK;
  listp->rule = rule;
  listp->link = NULL;
  if (EC_output[ec] == NULL)
    EC_output[ec] = listp;
  else
    EC_output_tail[ec]->link = listp;
  EC_output_tail[ec] = listp;

  return (listp);
}


/* print eclr_violation  */

static
void
eclr_violation(states)
     int states;
{
  fprintf(stderr, "%s: %d ECLR violation%s on %d state%s.\n",
	  infile, nviolation, (nviolation == 1) ? "":"s",
	  states, (nviolation == 1) ? "":"s");

  fprintf(stderr, "%s: please generate \"%s\" with --verbose option and see it.\n", infile, outfile);
  
  /* exit status  */
  failure = 1;
}


/* free rules assinged in 'EC_expanded[]' if any. and set all elements
   to NULL.
   and also reset all bits in 'EC_copied_pure'.  */

static
void
clear_EC()
{
  register int i;
  register semantic_rule_list *lp, *next;
  
  for (i=0; i < nEC; i++)
    {
      if (EC_expanded[i])
	{
	  free_semrules_pure(EC_expanded[i]);
	  EC_expanded[i] = NULL;
	}

      EC_to_write[i] = NULL;
      EC_rule[i] = NULL;
      if (verboseflag && !simple_verboseflag)
	{
	  lp = EC_output[i];
	  while (lp)
	    {
	      next = lp->link;
	      FREE(lp);
	      lp = next;
	    }
	  EC_output[i] = EC_output_tail[i] = NULL;
	}
    }
}


static
void
clear_local()
{
  register semantic_rule_list *lp, *next;
  
  lp = local_to_write;
  while (lp)
    {
      next = lp->link;
      free_semrule_pure(lp->rule);
      lp = next;
    }
  local_to_write = local_to_write_last = NULL;

  lp = local_output;
  while (lp)
    {
      next = lp->link;
      FREE(lp);
      lp = next;
    }
  local_output = local_output_last = NULL;
}



static
void
free_semexps_pure(exp)
     semantic_expression *exp;
{
  register semantic_expression *ep, *next;

  ep = exp;
  while (ep)
    {
      next = ep->next;
      
      if (ep->kind == KINHERIT || ep->kind == KSYNTHESIZE ||
	  ep->kind == KLOCAL)
	FREE(ep->pointer.attribute);
      FREE(ep);

      ep = next;
    }
}



static
void
free_semexp_pure(exp)
     semantic_expression *exp;
{
  if (exp->kind == KINHERIT || exp->kind == KSYNTHESIZE || exp->kind == KLOCAL)
    FREE(exp->pointer.attribute);
  FREE(exp);
}



/* free all semantic rules.  */

static
void
free_semrules_pure(rules)
     semantic_rule *rules;
{
  register semantic_rule *next;

  while (rules)
    {
      next = rules->link;
      /* codes below are the same code as free_semrule_pure().
	 but to call free_semrule_pure() is too heavy.  */
      if (rules->expression)
	free_semexps_pure(rules->expression);
      FREE(rules);
      rules = next;
    }
}



/* free a semantic rule.  */

static
void
free_semrule_pure(rule)
     semantic_rule *rule;
{
  if (rule->expression)
    free_semexps_pure(rule->expression);
  FREE(rule);
}



/*
  copy one semantic rule.

  But listed belows are not purely copied. Another words, they are shared.
  More to say, changes for copied affect original contents implicitly.
  You should not modify these contents unless you know what happens after.

	*rule->expression->pointer
	*rule->expression->link->pointer
	... and so on. (the contents of which each 'pointer' member indicates)
*/

static
semantic_rule *
copy_semrules_pure(rule)
     semantic_rule *rule;
{
  register semantic_rule *rp;
  
  rp = NEW(semantic_rule);
  rp->link = NULL;
  rp->lineno = rule->lineno;
  rp->attribute = rule->attribute;
  rp->expression = copy_semexps_pure(rule->expression);
  
  return (rp);
}
  

/*
  copy expressions.

  But the contens of which each 'pointer' member indicates are shared.
  Which means that changes for the copied affect the originals.
  You should not modify the contents unless you know what happens after.
*/

static
semantic_expression *
copy_semexps_pure(exp)
     semantic_expression *exp;
{
  register semantic_expression *head, *tail, *ep;
  register attribute *ap, *ap2;

  head = tail = NULL;

  while (exp)
    {
      ep = NEW(semantic_expression);

      /* set 'pointer' member  */
      if (exp->kind == KINHERIT ||  exp->kind == KSYNTHESIZE ||
	  exp->kind == KLOCAL)	
	{
	  ap = exp->pointer.attribute;
	  
	  ap2 = NEW(attribute);
	  ap2->tag = ap->tag;
	  ap2->offset = ap->offset;
	  ap2->pure = ap->pure;

	  ep->pointer.attribute = ap2;
	}            
      else
	ep->pointer = exp->pointer;
      
      ep->kind = exp->kind;
      ep->next = exp->next;

      if (head)
	tail->next = ep;
      else
	head = ep;
      tail = ep;
      
      exp = exp->next;
    }	/* while  */

  return (head);
}


#ifdef DEBUG

static
void
print_item(itemno)
     int itemno;
{
  register short *sp, *dot;
  register int ruleno;
  
  sp = dot = &ritem[itemno];
  while (*sp > 0)
    sp++;

  ruleno = -*sp;

  printf("%s -> ", tags[rlhs[ruleno]]);

  sp = &ritem[rrhs[ruleno]];
  while(*sp > 0)
    {
      if (sp == dot)
	printf(" .");
      printf(" %s", tags[*sp++]);
    }
  if (sp == dot)
    printf(" .");

  printf(" (rule %d)\n", ruleno);
}

static
void
print_EC_expanded()
{
  register int i;

  for (i=0; i < nEC; i++)
    {
      printf("\tEC_expanded[%d] : ", i);
      if (EC_expanded[i])
	print_semrule_d(EC_expanded[i]);
      else
	printf("\n");
    }
}


#endif
