/* Semantic rule reader
   Copyright (C) 1991-1993 Masataka Sassa, Kazuhiro Kuroishi, Teruhisa Hirai.
   Copyright (C)      1994,1995                              Yoshiki Ohshima.
     
This file is part of Rie, A Compiler Compiler Based on ECLR-attributed Grammar.

Rie is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

Rie is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bison; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

Bug reports, fixes or ideas are welcomed.
		Address: Yoshiki Ohshima
			 Tokyo Institute of Technology,
			 Department of Information Science,
			 Ookayama, Meguro-ku, Tokyo, 152 Japan.

		E-mail: rie-comments@is.titech.ac.jp			*/

#include <stdio.h>
#include "ctype.h"
#include "attribute.h"
#include "error.h"
#include "files.h"
#include "gram.h"
#include "lex.h"
#include "machine.h"
#include "new.h"
#include "pure.h"
#include "stringmgr.h"
#include "symtab.h"

/* See the description about following variables in reader.c  */
extern semantic_rule *first_semantic_rule;	/* reader.c  */
extern semantic_rule *last_semantic_rule;	/* reader.c  */
extern semantic_condition *first_condition;	/* reader.c  */
extern semantic_condition *last_condition; 	/* reader.c  */
extern semantic_rule **current_semantic_rules_p;/* reader.c  */
extern production_rule *current_prule;		/* reader.c  */
extern production_rule *current_prule_tail;	/* reader.c  */

extern int lineno;			/* reader.c  */
extern char *token_buffer;		/* reader.c  */

#if (defined(__STDC__) || defined(__GNUC__))

extern void fatal(char *);		/* main.c  */
extern void fatals(char *,...);		/* main.c  */
extern void init_lex_for_semrule(void);	/* lex.c  */
extern int lex_for_semrule(void);	/* lex.c  */
extern void unlex_for_semrule(int);	/* lex.c  */
extern void discard_semrule(void);	/* lex.c  */
extern void send_back_token(void);	/* lex.c  */
extern int lex_for_C_block(void);	/* lex.c  */
extern void check_well_defined(void);	/* check.c  */
extern void check_L_attributed(void);	/* check.c  */
extern void check_unused_local(void);	/* check.c  */

#ifdef DEBUG
extern void print_semrules_d(semantic_rule *); /* attribute.c  */
#endif /* DEBUG */

#else	/* not ANSI C nor GNU C  */

extern void fatal();			/* main.c  */
extern void fatals();			/* main.c  */
extern void init_lex_for_semrule();	/* lex.c  */
extern int lex_for_semrule();		/* lex.c  */
extern void unlex_for_semrule();	/* lex.c  */
extern void discard_semrule();		/* lex.c  */
extern void send_back_token();		/* lex.c  */
extern int lex_for_C_block();		/* lex.c  */
extern void check_well_defined();	/* check.c  */
extern void check_L_attributed();	/* check.c  */
extern void check_unused_local();	/* check.c  */

#ifdef DEBUG
extern void print_semrules_d(); /* attribute.c  */
#endif /* DEBUG */

#endif	/* nor ANSI C nor GNU C  */


#if (defined(__STDC__) || defined(__GNUC__))

void read_semantic_rules(int);
static void parse_assignment(void);
static void parse_local(void);
static int parse_local_assign(char *, char *);
static void parse_transfer(void);
static void parse_thread(void);
static void parse_except(void);
static void parse_condition(int);
static void parse_message(int);
static void parse_action(int);
static void parse_debug(int);
static attribute *parse_attribute(void);
static semantic_expression *parse_expressions(void);
static semantic_expression *parse_C_block(void);
static semantic_expression *allow_lhs_synthesized(semantic_expression *,int);
static void reassign(semantic_rule *);

static void adjust_offset(int);
static void set_local_offset(int);
       void append_to_current_prule(bucket *, int);
static attribute *make_attribute(char *, int, int, pure *);
static semantic_expression *make_expression(int, attribute *);
static semantic_expression *make_code(char *);
static semantic_condition *make_condition(int, semantic_expression *, int);
static void initialize_appeared_attributes(void);
static void free_appeared_attributes(void);
static attribute *append_local(char *, char *);
static void append_to_appeared_attributes(attribute *);
static void append_to_first_semantic_rule(semantic_expression *, attribute *);
       void append_terminator_to_first_semantic_rule(void);
static void append_to_first_condition(semantic_condition *);
       void append_terminator_to_first_condition(void);
static production_rule *search_prule(char *, int);
static attribute *search_attribute(production_rule *, char *);
static attribute *search_local(char *);
static attribute *search_attribute_in_appeared_attributes(attribute *);
       semantic_rule *search_rule_in_first_semantic_rule(attribute *);
       void remove_rule_from_first_semantic_rule(semantic_rule *);
static semantic_expression *append_to_semexp(semantic_expression *,semantic_expression *);
static void free_semexps(semantic_expression *);
static void remove_from_appeared_attributes(attribute *);
static void free_attribute_lists(attribute_list *);
       void free_current_prule(void);

#else	/* not ANSI C nor GNU C  */
void read_semantic_rules();
static void parse_assignment();
static void parse_local();
static int parse_local_assign();
static void parse_transfer();
static void parse_thread();
static void parse_except();
static void parse_condition();
static void parse_message();
static void parse_action();
static void parse_debug();
static attribute *parse_attribute();
static semantic_expression *parse_expressions();
static semantic_expression *parse_C_block();
static semantic_expression *allow_lhs_synthesized();
static void reassign();

static void adjust_offset();
static void set_local_offset();
       void append_to_current_prule();
static attribute *make_attribute();
static semantic_expression *make_expression();
static semantic_expression *make_code();
static semantic_condition *make_condition();
static attribute *append_local();
static void initialize_appeared_attributes();
static void free_appeared_attributes();
static void append_to_appeared_attributes();
static void append_to_first_semantic_rule();
       void append_terminator_to_first_semantic_rule();
static void append_to_first_condition();
       void append_terminator_to_first_condition();
static production_rule *search_prule();
static attribute *search_attribute();
static attribute *search_local();
static attribute *search_attribute_in_appeared_attributes();
       semantic_rule *search_rule_in_first_semantic_rule();
       void remove_rule_from_first_semantic_rule();
static semantic_expression *append_to_semexp();
static void free_semexps();
static void remove_from_appeared_attributes();
static void free_attribute_lists();
       void free_current_prule();

#endif	/* not ANSI C nor GNU C  */

attribute_list *appeared_attributes; /* attribute list for checking
				   if attribute is assigned or not.  */

#define TEMP_BUFFER_SIZE	40
char tmp_buffer[TEMP_BUFFER_SIZE];	/* share with check.c  */
#undef  TEMP_BUFFER_SIZE

/* parse semantic rules.  */

void
read_semantic_rules(right_length)
     int right_length;
{
  register int token;
  register int lhs_offset;   /* offset of lhs of production rule */
  register int base_offset;  /* base offset for outputing message and so on. */

  initialize_appeared_attributes();

  base_offset = nitems -1;
  lhs_offset = base_offset - right_length;

  init_lex_for_semrule();
  adjust_offset(right_length);

  for (;;)
    {
      token = lex_for_semrule();

      switch (token)
	{
	case IDENTIFIER:
	  parse_assignment();
	  break;

	case LOCALS:
	  parse_local();
	  break;

	case TRANSFER:
	  parse_transfer();
	  break;

	case THREAD:
	  parse_thread();
	  break;

	case EXCEPT:
	  parse_except();
	  break;

	case CONDITION:
	  parse_condition(base_offset);
	  break;

	case MESSAGE:
	  parse_message(base_offset);
	  break;

	case ACTION:
	  parse_action(base_offset);
	  break;

	case DEBUGS:
	  parse_debug(base_offset);
	  break;
	  
	case RIGHT_CURLY:
	  /* calculate and set offsets of local attributes.  */
	  set_local_offset(lhs_offset);

	  /* check if the grammer is Well-Defined or not.  */
	  check_well_defined();

	  /* check if the grammer is L-attributed or not.  */
	  check_L_attributed();

	  /* check unused local attribute.  */
	  check_unused_local();

	  /* free the 'appeared_attributes'.  */
	  free_appeared_attributes();

	  return;

	case SEMICOLON:
	  continue;

	case ENDFILE:
	  fatal("unterminated semantic rule.");

	default:
	  recovery("invalid character is found in semantic rule description.");
	  discard_semrule();
	}	/* switch */
    }	/* for */
}



/* parse a normal semantic rule.  */

static
void
parse_assignment()
{
  register int token;
  register attribute *attrp;
  register semantic_expression *exp;
  register semantic_rule *rule;
  
#ifdef DEBUG1
  printf("\nparse assignment rule following \"%s\"...\n", token_buffer);
#endif

  attrp = search_local(token_buffer);

  if (attrp)
    {
      /* this is a local attribute.  */
      rule = search_rule_in_first_semantic_rule(attrp);

      if (rule)
	{
	  recovery("local attribute \"%s\" is reassigned;"
		    , attrp->pure->name);
	  recovery_with_line(rule->lineno, "it is previously assigned here.");
	  discard_semrule();
	  return;
	}

      token = lex_for_semrule();

      if (token != ASSIGN)
	{
	  recovery("missing '=' following local attribute \"%s\"."
		    , attrp->pure->name);
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}

      exp = parse_expressions();

      if (exp == NULL)
	{
	  recovery("require semantic expressions following \"%s\"."
		    , attrp->pure->name);
	  discard_semrule();
	  return;
	}
    }

  else	/* attrp == NULL  */
    {
      /* except local attribute.  */
      attrp = parse_attribute();

      if (attrp == NULL)
	{
	  /* no need message */
	  discard_semrule();
	  return;
	}

      rule = search_rule_in_first_semantic_rule(attrp);
      
      if (rule)
	{
	  reassign(rule);
	  discard_semrule();
	  return;
	}

      if (search_attribute_in_appeared_attributes(attrp))
	remove_from_appeared_attributes(attrp);

      token = lex_for_semrule();

      if (token != ASSIGN)
	{
	  if (attrp->index == 1)
	    tmp_buffer[0] = '\0';
	  else
	    sprintf(tmp_buffer, "[%d]", attrp->index);
	  recovery("missing '=' following \"%s%s.%s\"."
		   , attrp->tag, tmp_buffer, attrp->pure->name);
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}

      exp = parse_expressions();

      if (exp == NULL)
	{
	  if (attrp->index == 1)
	    tmp_buffer[0] = '\0';
	  else
	    sprintf(tmp_buffer, "[%d]", attrp->index);
	  recovery("require semantic expressions following \"%s%s.%s =\"."
		   , attrp->tag, tmp_buffer, attrp->pure->name);
	  discard_semrule();
	  return;
	}
    }				/* if..else  */

  token = lex_for_semrule();

  if (token != SEMICOLON)
    {
      recovery("missing ';' at the end of assignment statement.");
      unlex_for_semrule(token);
      discard_semrule();
    }

  append_to_first_semantic_rule(exp, attrp);

#ifdef DEBUG1
  printf("finish parsing assignment rule.\n");
#endif
}



/* parse %local rule.  */

static
void
parse_local()
{
  register int token;
  register char *sourcep, *destp, *local_name;
  char type_string[MAXTOKEN + 1];

#ifdef DEBUG1
  printf("\nparse %%local rule...\n");
#endif
  
  local_name = destp = type_string;

  token = lex_for_semrule();

  /*
     'type_string' includes 2 strings.
     The first is the characters from 'type_string' to 'local_name - 1'.
     Which holds the type name of the local attribute.
     The second is the characters follwing 'local_name', which holds the
     local attribute name.
  */
  while (token == IDENTIFIER)
    {
      /* save 1 token of attribute type into type_string.
	 ex. "int" => "int", "struct foo" => "struct foo" */
      local_name = destp;
      sourcep = token_buffer;

      while(*sourcep && destp < type_string + MAXTOKEN)
	*destp++ = *sourcep++;
      *destp++ = ' ';

      token = lex_for_semrule();
    }

  unlex_for_semrule(token);

  if (destp == type_string)
    {
      recovery("require type name following %%local.");
      discard_semrule();
      return;
    }

  if (local_name == type_string)
    {
      recovery("require both type name and local attribute name following %%local.");
      discard_semrule();
      return;
    }
      
  *(local_name - 1) = '\0';
  *--destp	    = '\0';
  
  token = parse_local_assign(type_string, local_name);

  for (;;)
    {
      switch (token)
	{
	case COMMA:
	  token = lex_for_semrule();
	  continue;

	case SEMICOLON:
#ifdef DEBUG1
	  printf("finish parsing %%local rule.\n");
#endif
	  return;

	case IDENTIFIER:
	  token = parse_local_assign(type_string, token_buffer);
	  break;

	default:
	  recovery("invalid character is found in %%local.");
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}
    }
}



/* parse following decralation of local attribute.
   return the next token. called from parse_local().  */

static
int
parse_local_assign(type, name)
     char *type;
     char *name;
{
  register attribute *attrp;
  register semantic_expression *exp;
  register int token;
  
  attrp = search_local(name);	

  if (attrp != NULL)
    {
      warning("local attribute \"%s\" is redeclared.", name);
      discard_semrule();
      /* next to finish parsing %local.  */
      return (SEMICOLON);
    }

  attrp = append_local(name, type);

  token = lex_for_semrule();

  if (token == ASSIGN)
    {
      exp = parse_expressions();
      if (exp == NULL)
	{
	  recovery("require semantic expressions following \"%s =\"."
		    , attrp->tag);
	  discard_semrule();
	  /* next to finish parsing %local.  */
	  return (SEMICOLON);
	}
      append_to_first_semantic_rule(exp, attrp);
      token = lex_for_semrule();
    }

  if (token != SEMICOLON && token != COMMA)
    {
      recovery("missing ',' or ';' at the end of %%local.");
      unlex_for_semrule(token);
      discard_semrule();
      /* next to finish parsing %local.  */
      return (SEMICOLON);
    }

  return (token);
}



/* parse %transfer rule. */

static
void
parse_transfer()
{
  register int token;
  register int i = 0;
  register production_rule *prule;
  register attribute *attrp, *attrp2, *attrp3;
  register semantic_expression *exp;
  register semantic_rule *rule;

#ifdef DEBUG1
  printf("\nparse %%transfer rule...\n");
#endif	    

  for (;;)
    {
      token = lex_for_semrule();

      switch (token)
        {
	case COMMA:
	  continue;

	case SEMICOLON:
#ifdef DEBUG1
	  printf("finish parsing %%transfer rule.\n");
#endif
	  return;

	case IDENTIFIER:
	  attrp = search_attribute(current_prule, token_buffer);

	  if (attrp == NULL)
	    {          
	      recovery("lhs symbol doesn't have attribute \"%s\"."
		       , token_buffer);
	      break;	/* switch  */
	    }

	  exp = NULL;
	  prule = current_prule->next;

	  if (attrp->pure->class == INHERIT)
	    {
	      /* transfer inherited attributes from lhs to rhs.  */
	      for (; prule; prule = prule->next)
		{
		  attrp2 = search_attribute(prule, token_buffer);
		  if (attrp2 == NULL || attrp2->pure->class != INHERIT)
		    continue;	/* for  */

		  rule = search_rule_in_first_semantic_rule(attrp2);
		  if (rule)
		    {
		      reassign(rule);
		      /* dummy assign  */
		      exp = rule->expression;
		    }
		  else if (search_attribute_in_appeared_attributes(attrp2) == NULL)
		    {
		      /* make a rule  */
		      append_to_appeared_attributes(attrp2);
		      exp = make_expression(INHERIT, attrp);
		      append_to_first_semantic_rule(exp, attrp2);
		    }
		  /* through if 'attrp2' is declared as %except.  */
		}	/* for  */
	    }
	  else	/* attrp->pure->class != INHERIT (i.e. SYNTHESIZE)  */ 
	    {
	      /* transfer a synthesized attribute from one of rhs to lhs.  */
	      /* changed by OHSHIMA Yoshiki (1994/ 8/23) */
	      /* and made more change by OHSHIMA Yoshiki (1995/ 8/16) */
	      i = 0;
	      for (; prule; prule = prule->next)
		{
		  attrp2 = search_attribute(prule, token_buffer);
		  if (attrp2 != NULL && attrp2->pure->class == SYNTHESIZE)
		    {
			i++;
			attrp3 = attrp2;
			/* attrp3 keeps last attribute 'token_buffer'  in the prule. */
		    }
		}
	      switch (i)
		{
		  case 0:
		    break;
		  case 1:
		    rule = search_rule_in_first_semantic_rule(attrp);
		    if (rule)
		      {
			reassign(rule);
			/* dummy assign  */
			exp = rule->expression;
		      }
		    else if (search_attribute_in_appeared_attributes(attrp) == NULL)
		      {
			/* make a rule  */
			append_to_appeared_attributes(attrp);
			exp = make_expression(SYNTHESIZE, attrp3);
			append_to_first_semantic_rule(exp, attrp);
		      }
		    /* through if 'attrp3' is declared as %except.  */
		    break;
		  default:
		    /* this attribute appears more than two times in this production rule.  */
		    if (attrp3->index == 1)
		      tmp_buffer[0] = '\0';
		    else
		      sprintf(tmp_buffer, "[%d]", attrp3->index);
		      
		    warning("this transfer rule for attribute \"%s%s.%s\" has too many source attributes."
			    , attrp3->tag, tmp_buffer, token_buffer);

		    break;
		}
	    }
	  
	  if (exp == NULL)
	    /* this attribute wasn't transferred  */
	    recovery("no rhs symbol have attribute \"%s\"."
		      , token_buffer);
	  break;

	default:
	  recovery("missing ';' at the end of %%transfer.");
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}	/* switch  */
    }	/* for  */
}



/* parse %thread rule. */

static
void
parse_thread()
{
  register int token;
  register char *iname;   /* inherit attribute name */
  register char *sname;   /* synthesized attribute name */
  register semantic_expression *exp;
  register attribute *lattrp, *rattrp;
  register production_rule *prule;
  register semantic_rule *rule;
  
#ifdef DEBUG1
  printf("\nparse %%thread rule...\n");
#endif	    
  for (;;)
    {
      token = lex_for_semrule();

      switch (token)
        {
	case COMMA:
	  continue;

	case SEMICOLON:
#ifdef DEBUG1
	  printf("finish parsing %%thread rule.\n");
#endif
	  return;

	case IDENTIFIER:
	  exp = NULL;
	  rattrp = lattrp = NULL;
	  iname = copy_string(token_buffer);
	  token = lex_for_semrule();

	  if (token != IDENTIFIER)
	    {
	      recovery("inherit and synthesyze attributes pair is expected in %%thread.");
	      unlex_for_semrule(token);
	      discard_semrule();
	      return;
	    }

	  sname = copy_string(token_buffer);

	  /* 'iname' should be an inherited attribute.  */
	  prule = current_prule;
	  while (prule)
	    {
	      lattrp = search_attribute(prule, iname);
	      if (lattrp && lattrp->pure->class == INHERIT)
		break;
	      prule = prule->next;
	    }
	  if (prule == NULL)
	    {
	      recovery("no symbol have inherited attribute \"%s\".", iname);
	      break;	/* switch  */
	    }

	  /* 'sname' should be a synthesized attribute.  */
	  prule = current_prule;
	  while (prule)
	    {
	      lattrp = search_attribute(prule, sname);
	      if (lattrp && lattrp->pure->class == SYNTHESIZE)
		break;
	      prule = prule->next;
	    }
	  if (prule == NULL)
	    {
	      recovery("no symbol have synthesize attribute \"%s\".", sname);
	      break;	/* switch  */
	    }


	  /* is 'iname' an inherited attribute of lhs symbol ?  */
	  rattrp = search_attribute(current_prule, iname);
	  if (rattrp && rattrp->pure->class != INHERIT)
	    {
	      /* index of 'rattrp' is 1.  */
	      warning("\"%s.%s\" is not an inherit attribute."
		       , rattrp->tag, iname);
	      rattrp = NULL;
	    }

	  /* for each rhs symbols  */
	  prule = current_prule->next;
	  for (; prule; prule = prule->next)
	    {
	      /* pseudo loop  */
	      do	
		{
		  if (rattrp == NULL)
		    break;	/* pseudo loop  */
		  
		  /* source attribute is defined  */
		  lattrp = search_attribute(prule, iname);
		  if (lattrp == NULL)
		    break;	/* pseudo loop  */
		  
		  /* a destination inherited attribute is present.  */
		  if (lattrp->pure->class != INHERIT)
		    {
		      if (lattrp->index == 1)
			tmp_buffer[0] = '\0';
		      else
			sprintf(tmp_buffer, "[%d]", lattrp->index);
		      warning("\"%s%s.%s\" is not an inherit attribute."
			       , lattrp->tag, tmp_buffer, iname);
		      rattrp = NULL;
		      break;	/* pseudo loop  */
		    }

		  rule = search_rule_in_first_semantic_rule(lattrp);
		  if (rule)
		    reassign(rule);
		  else if (search_attribute_in_appeared_attributes(lattrp) == NULL)
		    {
		      append_to_appeared_attributes(lattrp);
		      exp = make_expression(rattrp->pure->class, rattrp);
		      append_to_first_semantic_rule(exp, lattrp);
		    }
		  /* through if 'lattrp' is declared as %except.  */
		}
	      while (0);	/* pseudo loop  */

	      lattrp = search_attribute(prule, sname);

	      if (lattrp)
		{
		  /* this symbol has a source synthesized attribute.  */
		  if (lattrp->pure->class != SYNTHESIZE)
		    {
		      if (lattrp->index == 1)
			tmp_buffer[0] = '\0';
		      else
			sprintf(tmp_buffer, "[%d]", lattrp->index);
		      warning("\"%s%s.%s\" is not a synthesized attribute."
			       , lattrp->tag, tmp_buffer, sname);
		      rattrp = NULL;
		      continue;		/* for  */
		    }
		  rattrp = lattrp;
		}
	    }	/* for  */

	  /* pseudo loop  */
	  do 
	    {
	      if (rattrp == NULL)
		break;		/* pseudo loop  */

	      lattrp = search_attribute(current_prule, sname);

	      if (lattrp == NULL)
		break;		/* pseudo loop  */
	      
	      if (lattrp->pure->class != SYNTHESIZE)
		{
		  /* index of 'lattrp' is 1.  */
		  warning("\"%s.%s\" is not a synthesized attribute."
			   , lattrp->tag, sname);
		  break;	/* pseudo loop  */
		}

	      rule = search_rule_in_first_semantic_rule(lattrp);

	      if (rule)
		reassign(rule);
	      else if (search_attribute_in_appeared_attributes(lattrp) == NULL)
		{
		  append_to_appeared_attributes(lattrp);
		  exp = make_expression(rattrp->pure->class, rattrp);
		  append_to_first_semantic_rule(exp, lattrp);
		}
	      /* through if 'lattrp' is declared as %except.  */
	    }
	  while(0);
	  
	  if (exp == NULL)
	    warning("%%thread %s %s generates no semantic rule."
		     , iname, sname);
	  break;

	default:
	  recovery("missing ';' at the end of %%thread.");
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}	/* switch  */
    }	/* for  */
}



/* parse %except rule.  */

static
void
parse_except()
{
  register int token;
  register attribute *attrp;
  register semantic_expression *exp;
  register semantic_rule *rule;

#ifdef DEBUG1
  printf("\nparse %%except rule...\n");
#endif

  token = lex_for_semrule();

  for (;;)
    {
      switch (token)
	{
	case COMMA:
	  token = lex_for_semrule();
	  continue;

	case SEMICOLON:
#ifdef DEBUG1
	  printf("finish parsing %%except rule.\n");
#endif
	  return;

	case IDENTIFIER:
	  attrp = parse_attribute();

	  if (attrp == NULL)
	    {
	      /* no need message */
	      discard_semrule();
	      return;
	    }
	      
	  rule = search_rule_in_first_semantic_rule(attrp);
	  attrp = search_attribute_in_appeared_attributes(attrp);

	  if (rule && attrp == NULL)
	    {
	      /* attribute is explicitly assigned.
		 not by %transfer nor %thread.  */
	      reassign(rule);
	      discard_semrule();
	      return;
	    }

	  if (attrp && rule == NULL)
	    {
	      if (attrp->index == 1)
		tmp_buffer[0] = '\0';
	      else
		sprintf(tmp_buffer, "[%d]", attrp->index);

	      warning("\"%s%s.%s\" is redeclared in %%except."
		      , attrp->tag, tmp_buffer, attrp->pure->name);
	      discard_semrule();
	      return;
	    }

	  if (rule)
	    /* cancel previous assignment, which is done by %thread
	       or %transfer.  */
	    remove_rule_from_first_semantic_rule(rule);
	  else
	    /* except declaration.  */
	    append_to_appeared_attributes(attrp);

	  token = lex_for_semrule();

	  if (token == ASSIGN)
	    {
	      remove_from_appeared_attributes(attrp);

	      exp = parse_expressions();
	      if (exp == NULL)
		{
		  if (attrp->index == 1)
		    tmp_buffer[0] = '\0';
		  else
		    sprintf(tmp_buffer, "[%d]", attrp->index);
		  recovery("require semantic expressions following \"%s%s.%s =\"."
			    , attrp->tag, tmp_buffer, attrp->pure->name);
		  discard_semrule();
		  return;
		}

	      append_to_first_semantic_rule(exp, attrp);
	      token = lex_for_semrule();
	    }

	  if (token != SEMICOLON && token != COMMA)
	    {
	      recovery("missing ',' or ';' at the end of %%except.");
	      unlex_for_semrule(token);
	      discard_semrule();
	      return;
	    }
	  break;

	default:
	  recovery("invalid text in %%except.");
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}	/* switch  */
    }	/* for  */
}



/* parse %condition rule.  */

static
void
parse_condition(base_offset)
     int base_offset;
{
  register int token;
  register semantic_expression *exp;
  register semantic_condition *condp;

#ifdef DEBUG1
  printf("\nparse %%condition rule...\n");
#endif

  exp = parse_expressions();

  if (exp == NULL)
    {
      recovery("no conditional expressions following %%condition.");
      discard_semrule();
      return;
    }

  exp = allow_lhs_synthesized(exp, base_offset);

  condp = make_condition(CCONDITION, exp, base_offset);

  append_to_first_condition(condp);

  token = lex_for_semrule();

  if (token == SEMICOLON)
    token = lex_for_semrule();

  switch (token)
    {
    case MESSAGE:
      parse_message(base_offset);
      break;

    case ACTION:
      parse_action(base_offset);
      break;

    default:
      recovery("require %%message or %%action following %%condition.");
      unlex_for_semrule(token);
      discard_semrule();
      break;
    }
#ifdef DEBUG1
  printf("finish parsing %%condition rule.\n");
#endif
}



/* Parse %message rule.  */

static
void
parse_message(base_offset)
     int base_offset;
{
  register int token, last_token;
  register semantic_expression *exp;

#ifdef DEBUG1
  printf("\nparse %%message rule...\n");
#endif

  exp = parse_expressions();

  if (exp == NULL)
    {
      recovery("no format string following %%message.");
      discard_semrule();
      return;
    }

  exp = allow_lhs_synthesized(exp, base_offset);

  last_token = MESSAGE;		/* for safety  */
  for (;; last_token = token)
    {
      semantic_expression *tmpexp;
      semantic_condition *condp;
      
      token = lex_for_semrule();
      
      switch (token)
	{
	case SEMICOLON:
	  condp = make_condition(CMESSAGE, exp, base_offset);
	  append_to_first_condition(condp);
#ifdef DEBUG1
	  printf("finish parsing %%message rule.\n");
#endif
	  return;

	case COMMA:
	  if (last_token == COMMA)
	    {
	      recovery("',' follows ',' at %%message argument.\n");
	      discard_semrule();
	      return;
	    }
	  tmpexp = make_code(token_buffer);
	  exp = append_to_semexp(tmpexp, exp);
	  break;

	case STRING:		/* string constants  */
	case IDENTIFIER:	/* a C variable or preprocessor macro  */
	case CODES:
	  unlex_for_semrule(token);
	  tmpexp = parse_expressions();
	  if (tmpexp == NULL)
	    {
	      discard_semrule();
	      return;
	    }
	  tmpexp = allow_lhs_synthesized(tmpexp, base_offset);
	  exp = append_to_semexp(tmpexp, exp);
	  break;
	  
	default:
	  recovery("missing ';' at the end of %%message.");
	  unlex_for_semrule(token);
	  discard_semrule();
	  return;
	}
    }
}



/* parse %action rule.  */

static
void
parse_action(base_offset)
     int base_offset;
{
  register int token;
  register semantic_expression *exp, *tmpexp;
  register semantic_condition *condp;

#ifdef DEBUG1
  printf("\nparse %%action rule...\n");
#endif

  exp = NULL;
  token = lex_for_semrule();

  switch (token)
    {
    case LEFT_CURLY:
      exp = parse_C_block();
      exp = allow_lhs_synthesized(exp, base_offset);
      break;

    case CODES:
    case IDENTIFIER:
      unlex_for_semrule(token);
      exp = parse_expressions();
      if (exp == NULL)
	{
	  discard_semrule();
	  return;
	}
      exp = allow_lhs_synthesized(exp, base_offset);
      
      token = lex_for_semrule();
      if (token != SEMICOLON)
	{
	  recovery("missing ';' at the end of %%action");
	  unlex_for_semrule(token);
	  discard_semrule();
	}
      /* through. now "token == SEMICOLON"  */

    case SEMICOLON:
      tmpexp = make_code(token_buffer);
      exp = append_to_semexp(tmpexp, exp);
      break;
      
    default:
      recovery("invalid character is found in %%action");
      unlex_for_semrule(token);
      discard_semrule();
      break;
    }

  condp = make_condition(CACTION, exp, base_offset);
  append_to_first_condition(condp);
#ifdef DEBUG1
  printf("finish parsing %%action rule.\n");
#endif
}



/* parse %debug rule.  */

static
void
parse_debug(base_offset)
     int base_offset;
{
  register int token;
  register semantic_expression *exp, *tmpexp;
  register semantic_condition *condp;

#ifdef DEBUG1
  printf("\nparse %%debug rule...\n");
#endif

  exp = NULL;
  token = lex_for_semrule();

  switch (token)
    {
    case LEFT_CURLY:
      exp = parse_C_block();
      exp = allow_lhs_synthesized(exp, base_offset);
      break;

    case CODES:
    case IDENTIFIER:
      unlex_for_semrule(token);
      exp = parse_expressions();
      if (exp == NULL)
	{
	  discard_semrule();
	  return;
	}
      exp = allow_lhs_synthesized(exp, base_offset);
      
      token = lex_for_semrule();
      if (token != SEMICOLON)
	{
	  recovery("missing ';' at the end of %%debug");
	  unlex_for_semrule(token);
	  discard_semrule();
	}
      /* through. now "token == SEMICOLON"  */

    case SEMICOLON:
      tmpexp = make_code(token_buffer);
      exp = append_to_semexp(tmpexp, exp);
      break;
      
    default:
      recovery("invalid character is found in %%debug");
      unlex_for_semrule(token);
      discard_semrule();
      break;
    }

  condp = make_condition(CDEBUG, exp, base_offset);
  append_to_first_condition(condp);
#ifdef DEBUG1
  printf("finish parsing %%debug rule.\n");
#endif
}



/* parse attribute such as 'exp[1].S_env'.  */

static
attribute *
parse_attribute()
{
  register int token;
  register int index;
  register char *symbol_name;
  register production_rule *prule;
  register attribute *attrp;

  extern int numval;		/* defined in lex_for_semrule()	 */

  attrp = NULL;
  symbol_name = copy_string(token_buffer);

  token = lex_for_semrule();

  if (token == INDEX)
    {
      /* there's notation is like "exp[1]".  */
      index = numval;
      sprintf(tmp_buffer, "[%d]", index);
      token = lex_for_semrule();
    }
  else
    {
      index = 1;
      tmp_buffer[0] = '\0';
    }

  prule = search_prule(symbol_name, index);

  if (prule == NULL)
    {
      recovery("\"%s%s\" does not appear in this rule."
	       , symbol_name, tmp_buffer);
      unlex_for_semrule(token);
      return (NULL);
    }

  if (token != PERIOD)
    {
      recovery("missing '.' following symbol \"%s%s\"."
	       , symbol_name, tmp_buffer);
      unlex_for_semrule(token);
      return (NULL);
    }

  token = lex_for_semrule();

  if (token != IDENTIFIER)
    {
      recovery("require attribute name following symbol \"%s%s\".",
		symbol_name, tmp_buffer);
      unlex_for_semrule(token);
      return (NULL);
    }

  attrp = search_attribute(prule, token_buffer);

  if (attrp == NULL)
    recovery("unknown attribute \"%s%s.%s\"."
	     , symbol_name, tmp_buffer, token_buffer);
  
  return (attrp);
}



/* parse right hand side of semantic rule.  */

static
semantic_expression *
parse_expressions()
{
  register int token;
  register int level;
  register attribute *attrp;
  register pure *purep;
  register semantic_expression *exp, *head, *tail;
  register production_rule *prule;

  level = 0;
  head = tail = NULL;

  for (;;)
    {
      token = lex_for_semrule();

      switch (token)
	{
	case IDENTIFIER:
	  /* local attribute or attribute or C identifier */
	  attrp = search_local(token_buffer);
	  if (attrp)
	    {
	      /* this is a local attribute.  */
	      if (search_rule_in_first_semantic_rule(attrp) == NULL)
		{
		  recovery("local attribute \"%s\" is referred before assigned."
			    , attrp->pure->name);
		  return (NULL);
		}
	      exp = make_expression(LOCAL, attrp);
	      break;	/* switch  */
	    }

	  /* no more local attribute  */
	  prule = search_prule(token_buffer, 1);
	  if (prule != NULL)
	    {
	      /* this is an attribute.  */
	      attrp = parse_attribute();
	      if (attrp == NULL)
		return (NULL);

	      purep = attrp->pure;
	      exp = make_expression(purep->class, attrp);
	      break;
	    }
	  /* C identifier is treated as CODES.  */
	  /* through */

	case CODES:
	case PERIOD:
	case ASSIGN:
	case STRING:
	case INDEX:
	  exp = make_code(token_buffer);
	  break;

	case LEFT_PAREN:
	  level++;
	  exp = make_code(token_buffer);
	  break;

	case RIGHT_PAREN:
	  level--;
	  exp = make_code(token_buffer);
	  break;

	case COMMA:
	case SEMICOLON:
	  if (level > 0)
	    {
	      exp = make_code(token_buffer);
	      break;
	    }

	  /* COMMA and SEMICOLON are treated as a terminator.  */
	  /* through */
		
	default:
	  unlex_for_semrule(token);
	  return (head);
	}	/* switch  */

      if (head == NULL)
	head = exp;
      else
	tail->next = exp;
      tail = exp;
    }	/* for  */
}


/* parse expressions in C block.
   now just after '{'  */

static
semantic_expression *
parse_C_block()
{
  int token;
  int level;
  attribute *attrp;
  pure *purep;
  semantic_expression *exp, *head, *tail;
  production_rule *prule;

  head = tail = make_code(token_buffer);

  level = 1;
  for (; 0<level;)
    {
      token = lex_for_C_block();

      switch (token)
	{
	case IDENTIFIER:
	  /* local attribute or attribute or C identifier */
	  attrp = search_local(token_buffer);
	  if (attrp)
	    {
	      /* this is a local attribute.  */
	      if (search_rule_in_first_semantic_rule(attrp) == NULL)
		{
		  recovery("local attribute \"%s\" is referred before assigned."
			    , attrp->pure->name);
		  return (NULL);
		}
	      exp = make_expression(LOCAL, attrp);
	      break;	/* switch  */
	    }

	  /* no more local attribute  */
	  prule = search_prule(token_buffer, 1);
	  if (prule)
	    {
	      /* this is an attribute.  */
	      attrp = parse_attribute(); /* ok to use parse_attribute() */
	      if (attrp == NULL)
		return (NULL);

	      purep = attrp->pure;
	      exp = make_expression(purep->class, attrp);
	      break;
	    }
	  /* C identifier is treated as CODES.  */
	  /* through */

	case CODES:
	case PERIOD:
	case INDEX:
	  exp = make_code(token_buffer);
	  break;

	case LEFT_CURLY:
	  level++;
	  exp = make_code(token_buffer);
	  break;

	case RIGHT_CURLY:
	  level--;
	  exp = make_code(token_buffer);
	  break;

	case EOF:
	  fatal("unteminated C block");

	default:
	  internal("at parse_C_block()");
	}	/* switch  */

      if (head == NULL)
	head = exp;
      else
	tail->next = exp;
      tail = exp;
    }	/* for  */

  return (head);
}


/* replace KSYNTHESIZE to KSYNTHESIZE_LHS if it is a synthesized attribute
   of the lhs. */

static
semantic_expression *
allow_lhs_synthesized(exps, base)
     semantic_expression *exps;
     int base;
{
  register attribute_list *alist;
  register attribute *ap;
  register semantic_expression *exp;

  exp = exps;
  while (exp)
    {
      if (exp->kind == KSYNTHESIZE && exp->pointer.attribute->offset == base)
	{
	  /* this expression may be a synthesized attribute of the lhs */

	  /* lhs attributes */
	  alist = current_prule->attribute_list;
	  ap = exp->pointer.attribute;
	  while (alist)
	    {
	      if (alist->attribute == ap)
		/* sure ! this is. */
		break;
	      alist = alist->link;
	    }

	  if (alist)
	    /* yes, this is the synthesized attribute of the lhs */
	    exp->kind = KSYNTHESIZE_LHS;
	}
      exp = exp->next;
    }

  return (exps);
}
  

/* print an error message for reassigning attribute.  */

static
void
reassign(rule)
     semantic_rule *rule;
{
  register attribute *attrp;

  attrp = rule->attribute;

  if (attrp == NULL)
    internal("at reassign()");
  
  if (attrp->index == 1)
    tmp_buffer[0] = '\0';
  else
    sprintf(tmp_buffer, "[%d]", attrp->index);
  warning("\"%s%s.%s\" is reassigned;"
	  , attrp->tag, tmp_buffer, attrp->pure->name);
  warning_with_line(rule->lineno, "it is previously assigned here.");
}



/* these above are main functions of parsing the semantic rule.  */

/* adjust a offset in the current production rule.  */

static
void
adjust_offset(rhs_length)
     int rhs_length;
{
  register production_rule *prule;
  register attribute *attrp;
  register attribute_list *alist;

  alist = current_prule->attribute_list;

  /* adjust offset of synthesized attribute of lhs symbol.  */
  while (alist)
    {
      attrp = alist->attribute;
      if (attrp->pure->class == SYNTHESIZE)
	attrp->offset += rhs_length;
      alist = alist->link;
    }

  prule = current_prule->next;

  /* adjust offsets of inherited attribute of each rhs symbol.  */
  while (prule)
    {
      alist = prule->attribute_list;

      while (alist)
	{
	  attrp = alist->attribute;
	  if (attrp->pure->class == INHERIT)
	    attrp->offset--;
	  alist = alist->link;
	}
      prule = prule->next;
    }
}



/* set offsets of local attributes.  */

static
void
set_local_offset(lhs_offset)
     int lhs_offset;
{
  register semantic_rule *rule;
  register semantic_expression *exp;
  register short maxoff;
  
  rule = *current_semantic_rules_p;

  while (rule)
    {
      if (rule->attribute->pure->class == LOCAL)
	{
	  maxoff = lhs_offset;

	  exp = rule->expression;
	  while (exp)
	    {
	      if ((exp->kind == KINHERIT || exp->kind == KSYNTHESIZE
		  || exp->kind == KLOCAL)
		  &&
		  exp->pointer.attribute->offset > maxoff)
		maxoff = exp->pointer.attribute->offset;
	      exp = exp->next;
	    }

	  rule->attribute->offset = maxoff;
	}
      rule = rule->link;
    }
}



/* make a list of production rule with attribute.  */


void
append_to_current_prule(symbol, offset)
     bucket *symbol;
     int offset;
{
  register pure_list *list;
  register production_rule *prule;
  register attribute_list *alist;
  register int index;
  register char *tag;

  tag = symbol->tag;

  /* get index from production rule.  */
  index = 1;
  prule = current_prule;
  while (prule)
    {
      /* IF YOU DON'T USE THE STRING MANAGER. USE 'strcmp()' INSTEAD.  */
      if (prule->tag == tag)
	index++;
      prule = prule->next;
    }
  
  prule = NEW(production_rule);
  prule->next = NULL;
  prule->tag = tag;

  list = symbol->pure_list;

  while (list)
    {
      /* add each attribute of the symbol to prule.  */
      alist = NEW(attribute_list);
      alist->attribute = make_attribute(tag, offset, index, list->pure);
      
      alist->link = prule->attribute_list;
      prule->attribute_list = alist;

      list = list->link;
    }

  if (current_prule == NULL)
    {
      current_prule = prule;
    }
  else
    current_prule_tail->next = prule;
  current_prule_tail = prule;
}



/* make an attribute.  */

static
attribute *
make_attribute(name, offset, index, purep)
     char *name;
     int offset;
     int index;
     pure *purep;
{
  register attribute *attrp;

  attrp = NEW(attribute);
  attrp->tag = name;
  attrp->offset = offset;
  attrp->index = index;
  attrp->pure = purep;

  return (attrp);
}



/* make an expression holding an attribute.  */

static
semantic_expression *
make_expression(class, attrp)
     short class;
     attribute *attrp;
{
  register semantic_expression *exp;

  exp = NEW(semantic_expression);
  exp->pointer.attribute = attrp;
  exp->next = NULL;

  /* set 'kind' member.  */
  switch (class)
    {
    case INHERIT:
      exp->kind = KINHERIT;
      break;

    case SYNTHESIZE:
      exp->kind = KSYNTHESIZE;
      break;

    case LOCAL:
      exp->kind = KLOCAL;
      break;

    default:
      /* 'CODE' are generated by make_code().  */
      exp->kind = KUNKNOWN;
      break;
    }
  
  return (exp);
}



/* make an expression holding a code.  */

static
semantic_expression *
make_code(code)
     char *code;
{
  register semantic_expression *exp;

  exp = NEW(semantic_expression);
  exp->kind = KCODE;
  exp->pointer.code = copy_string(code);
  exp->next = NULL;

  return (exp);
}



/* make a 'semantic_condition' type  */

static
semantic_condition *
make_condition(kind, exp, base_offset)
     int kind;
     semantic_expression *exp;
     int base_offset;
{
  register semantic_condition *condp;

  condp = NEW(semantic_condition);
  condp->kind = kind;
  condp->lineno = lineno;	/* file global variable */
  condp->expression = exp;
  condp->base_offset = base_offset;

  return (condp);
}


/* append 'name', whose type is 'type', into 'appeared_attributes' as a local
   attribute.  */

static
attribute *
append_local(name, type)
     char *name;
     char *type;
{
  register pure *purep;
  register attribute *attrp;
  char buffer[10];

#ifdef DEBUG1
  printf("append local attribute \"%s %s\"\n", type, name);
#endif
  
  purep = NEW(pure);
  purep->name = copy_string(name);
  purep->type = copy_string(type);
  purep->class = LOCAL;
  purep->EC = EC_UNDEF;

  /* symbol name of local attribute means rule number.  */
  sprintf(buffer, "%d", nrules);
  attrp = make_attribute(copy_string(buffer), UNDEF, UNDEF, purep);
  append_to_appeared_attributes(attrp);

  return (attrp);
}



/* initialize 'appeared_attributes'  */

static
void
initialize_appeared_attributes()
{
  appeared_attributes = NULL;
}


/* free memories of 'appeared_attributes'  */

static
void
free_appeared_attributes()
{
  free_attribute_lists(appeared_attributes);
  appeared_attributes = NULL;
}


/* append an attribute into 'appeared_attributes'.  */

static
void
append_to_appeared_attributes(attrp)
     attribute *attrp;
{
  register attribute_list *alist;

  alist = NEW(attribute_list);
  alist->attribute = attrp;
  alist->link = appeared_attributes;
  appeared_attributes = alist;
}



/* append a semantic rule into 'first_semantic_rule'.  */

static
void
append_to_first_semantic_rule(exp, attrp)
     semantic_expression *exp;
     attribute *attrp;
{
  register semantic_rule *rule;

  rule = NEW(semantic_rule);
  rule->link = NULL;
  rule->lineno = lineno;
  rule->expression = exp;
  rule->attribute = attrp;

  if (first_semantic_rule == NULL)
    first_semantic_rule = rule;
  else
    last_semantic_rule->link = rule;
  last_semantic_rule = rule;

#ifdef DEBUG1
  print_semrules_d(rule);
#endif
}



/* put a terminator to 'first_semantic_rule'  */

void
append_terminator_to_first_semantic_rule()
{
  append_to_first_semantic_rule(NULL, NULL);
  current_semantic_rules_p = &last_semantic_rule->link;
#ifdef DEBUG1
  printf("parsed\n");
#endif /* DEBUG1 */
}  



/* append a context condition to 'first_condition'  */

static
void
append_to_first_condition(condp)
     semantic_condition *condp;
{
  if (first_condition == NULL)
    first_condition = condp;
  else
    last_condition->link = condp;
  last_condition = condp;
}


void
append_terminator_to_first_condition()
{
  semantic_condition *condp;
  condp = make_condition(CTERMINATOR, NULL, -1234);
  append_to_first_condition(condp);
}


/* search a 'index'th grammar symbol 'name' in 'current_prule'.  */
   
static
production_rule *
search_prule(name, index)
     char *name;
     int index;
{
  register production_rule *prule;

  prule = current_prule;

  for (; prule; prule=prule->next)
    if (strcmp(name, prule->tag) == 0 && --index <= 0)
      break;

  return (prule);
}



/* search an attribute.  */

static
attribute *
search_attribute(prule, name)
     production_rule *prule;
     char *name;
{
  register attribute *attrp;
  register attribute_list *alist;

  alist = prule->attribute_list;
  
  while (alist)
    {
      attrp = alist->attribute;
      if (strcmp(name, attrp->pure->name) == 0)
        return (attrp);

      alist = alist->link;
    }

  return (NULL);
}



/* Search a local attribute in 'appeared_attributes'.  */

static
attribute *
search_local(name)
     char *name;
{
  register pure *purep;
  register attribute_list *alist;
  
  alist = appeared_attributes;

  while (alist)
    {
      purep = alist->attribute->pure;
      /* YOU NEVER USE "==". 'name' MAY NOT REGISTERED BY THE STRING MANAGER.*/
      if (purep->class == LOCAL && strcmp(purep->name, name) == 0)
	return (alist->attribute);
      alist = alist->link;
    }

  return (NULL);
}



/* Search an attribute in 'appeared_attributes'.  */

static
attribute *
search_attribute_in_appeared_attributes(attrp)
     attribute *attrp;
{
  register attribute_list *alist;

  alist = appeared_attributes;

  for (; alist; alist=alist->link)
    if (alist->attribute == attrp)
      break;

  if (alist)
    return (alist->attribute);
  else
    return (NULL);
}



/* Search an attribute in '*current_semantic_rules_p'.  */

semantic_rule *
search_rule_in_first_semantic_rule(attrp)
     attribute *attrp;
{
  register semantic_rule *rule;

  rule = *current_semantic_rules_p;

  for (; rule; rule=rule->link)
    if (rule->attribute == attrp)
      break;

  return (rule);
}



/* remove a rule from '*current_semantic_rules_p'.  */

void
remove_rule_from_first_semantic_rule(rule)
     semantic_rule *rule;
{
  semantic_rule *rp, *previous;

  rp = *current_semantic_rules_p;
  previous = NULL;

  for (; rp; previous=rp, rp=rp->link)
    {
      if (rp == rule)
	{
	  free_semexps(rule->expression);

	  if (previous)
	    previous->link = rp->link;
	  else
	    *current_semantic_rules_p = rp->link;
 
 	  if (rule == last_semantic_rule)
 	    last_semantic_rule = previous;

	  FREE(rule);		/* IS THIS OK ?  */

	  break;
	}
    }
}



/* remove an attribute from 'appeared_attributes'.  */

static
void
remove_from_appeared_attributes(attrp)
     attribute *attrp;
{
  register attribute_list *current, *previous;

  current = appeared_attributes;
  previous = NULL;

  while (current)
    {
      if (current->attribute == attrp)
	{
	  if (previous)
	    previous->link = current->link;
	  else
	    appeared_attributes = current->link;
	  FREE(current);

	  break;
	}

      previous = current;
      current = current->link;
    }
}


static
semantic_expression *
append_to_semexp(from, to)
     semantic_expression *from;
     semantic_expression *to;
{
  semantic_expression *first, *last;
  first = last = to;
  for (; to; to = to->next)
    last = to;
  if (last)
    {
      last->next = from;
      return (first);
    }
  else
    return (NULL);
}


/* free expressions  */

static
void
free_semexps(exp)
     semantic_expression *exp;
{
  register semantic_expression *next;

  while (exp)
    {
      next = exp->next;
      FREE(exp);
      exp = next;
    }
}



/* free attribute lists.  */

static
void
free_attribute_lists(alist)
     attribute_list *alist;
{
  register attribute_list *next;

  while (alist)
    {
      next = alist->link;
      FREE(alist);
      alist = next;
    }
}



/* free production rule list.  */

void
free_current_prule()
{
  while (current_prule)
    {
      production_rule *prule;

      free_attribute_lists(current_prule->attribute_list);
      prule = current_prule->next;
      FREE(current_prule);
      current_prule = prule;
    }
  current_prule_tail = NULL;	/* also 'current_prule == NULL'  */
}
